/***************************************************************************
 *   Copyright (C) 2006 by Dirk Hartmann                                   *
 *   dirk.hartmann@sitter-net.de                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <stdlib.h>
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>

#include "smartbeatclient.h"
#include "smartbeatlogger.h"
#include "config.h"

Smartbeatclient::Smartbeatclient(const std::string &configFile)
 : authed(0), mysqlReconnect(false)
{
	/* store configuration */
	config.readConfigFromFile(configFile);
	
	config.logger()->write_to_log("client started");
	initSocket();
	initGpg();
	connectToDatabases();
	mysqlStatus = "UNUSED";
}

void Smartbeatclient::initSocket()
{
	socket = new SocketClass(config.smartBeatServerAddress(), config.serverPort());
	socket->set_recv_timeout(config.socketTimout());
}

void Smartbeatclient::initGpg()
{
	/* init gpg */
	gpg.init(config.gpgCommand(), config.gpgHomeDir(), config.gpgKeyName());
}

void Smartbeatclient::connectToDatabases()
{
	const map<string, DatabaseSettings> *databaseSettings = config.databaseSettings(); // fetch from config
	map<string, DatabaseSettings>::const_iterator databaseIter;
	
	for (databaseIter = databaseSettings->begin(); databaseIter != databaseSettings->end();
		databaseIter++) {
		DatabaseSettings dbSettings = databaseIter->second;
		DBAccess *database = new DBAccess;
		database->init(dbSettings.dbHost, dbSettings.dbUser, dbSettings.dbPass, databaseIter->first);
		if (database->connectToDb() < 0) { // connect to database failed
			config.logger()->write_to_log("ERROR: Connect to database "+ databaseIter->first +" failed");
			config.logger()->write_to_log(database->read_error_message());
			delete database;
		} else {
			databases.insert(make_pair(databaseIter->first, database));
		}
	}
}

void Smartbeatclient::clearDatabases()
{
	map<string, DBAccess *>::iterator dbIter;
	for (dbIter = databases.begin(); dbIter != databases.end(); dbIter++) {
		dbIter->second->closeMysqlConnection();
		delete dbIter->second;
	}
	databases.clear();
}

void Smartbeatclient::reloadConfig()
{
	delete socket;
	clearDatabases();
	config.reloadConfig();
	initSocket();
	connectToDatabases();
	initGpg();
}

void Smartbeatclient::writeToLog(const std::string & msg)
{
	config.logger()->write_to_log(msg);
}


int Smartbeatclient::get_my_ip(std::string &ip)
{
	//TODO check if the returned vaule is in format ddd.ddd.ddd.ddd but programm will not crash when something else is returned :) handle ipv6?
	std::string msg;
	int first, last;
	int port = 80; // std http port
	std::string host;
	std::string ret = host = "";
	
	first = config.getMyIpUrl().find_first_not_of("http://",0);
	last = config.getMyIpUrl().find(":", first);
	/* no port specified in url */
	if (last == -1) {
		/* search for the next '/' */
		last = config.getMyIpUrl().find("/", first);
		if (last - first > 0)
			host = config.getMyIpUrl().substr(first, last-first);
	}
	else {
		if (last - first > 0)
			host = config.getMyIpUrl().substr(first, last-first);
		/* findout portnumber */
		first = last;
		/* port number is between : and / */
		last = config.getMyIpUrl().find("/", first);
		/* if found convert it to integer */
		if (last - first > 0)
			port = atoi(config.getMyIpUrl().substr(first, last-first).c_str());
	}
// 	std::cout << host << " " << port << std::endl;
	
	if (host != "") {
		SocketClass sock (host, port);
		if ( sock.connect() == -1) {
			config.logger()->write_to_log("ERROR: could not find out my ip, connect failed");
			config.logger()->write_to_log("ERROR: the reason: " +  sock.read_error());
			sock.close();
			return -1;
		}
		if (sock.send("GET " + config.getMyIpUrl() +"\r\n") == -1) {
			config.logger()->write_to_log("ERROR: could not find out my ip, get failed");
			config.logger()->write_to_log("ERROR: the reason: " +  sock.read_error());
			sock.close();
			return -1;
		}
		ip = "";
		sock.set_recv_timeout(5); // for preventing blocking
		sock.recieve(ip);
		sock.close();
		if (ip != "")
			return 1;
		else {
			config.logger()->write_to_log("ERROR: could not find out my ip, nothing returned");
			return -1;
		}
	}
	// example url:
	// http://www.sitter-net.de/cgi-bin/whatismyip.pl
	return 1;
}

std::string Smartbeatclient::find_out_diskspace(const std::string &path_to_partition) const
{
	std::string diskPercent, tmp;
	int first, last;

	diskPercent = "";

	/* run df -h */
	execute_command("df -h " + path_to_partition, tmp);
	
	/* find first '%' from right */
	last = tmp.find_last_of("%");
	
	/* max 99% allowed so 2 to left */
	first = last - 2;
	
	if (tmp[first] == ' ') // if less than 10 percent
		++first;
	else if (first > 0 and tmp[first-1] != ' ') // check for 100%
		--first;
	if (last - first > 0 and last > 0 and first > 0)
		diskPercent = tmp.substr(first, last-first);
	else
		config.logger()->write_to_log("WARNING: could not parse 'du -h'. check configfile for correct settings for partition " + path_to_partition);
	
	return diskPercent;
}

std::string Smartbeatclient::find_out_avg_load()
{
	std::string avg, tmp;
	int first, last;
	// float multiply_helper;

	avg = "";

	/* cat /proc/avgload to find out avg load */
	execute_command("cat /proc/loadavg", tmp);
	
	/* set first to 0 */
	first = 0;
	
	/* search fo first space */
	last = tmp.find(" ");
	avg = tmp.substr(first, last);
	
	return avg;
}

std::string Smartbeatclient::find_out_uptime( )
{
	std::string uptime, tmp;
	int first, last;

	uptime = ""; // init as empty

	/* cat /proc/uptime to find out uptime */
	execute_command("cat /proc/uptime", tmp);
	
	/* set first to 0 */
	first = 0;
	
	/* search fo first space */
	last = tmp.find(" ");
	uptime = tmp.substr(first, last);
	
	return uptime;
}

std::string Smartbeatclient::mix_message(const std::vector<std::string> &xmlTags) const
{
	std::vector<std::string> tags = xmlTags; // copy so that it can be modified
	int num, act_tags, size;
	std::string ret;
	
	if (tags.empty()) // preventing bugs
		return "";
	
	act_tags = tags.size();
	size = tags.size();
	ret = "";
	for (int i = 0; i < size; i++) {
		num = rand() % act_tags;
		ret += tags[num];
		tags.erase(tags.begin() + num);
		
		--act_tags;
	}
	return ret;
}

int Smartbeatclient::send_status(bool shortStatus)
{
	std::vector<std::string> tags;
	std::string load, uptime;

	/* monitored  disks */
	tags.push_back(tagsForMonitoredDisks());

	/* average load */
	load = find_out_avg_load();
	tags.push_back("<avgload>" + load + "</avgload>");
	
	/* uptime */
	uptime = find_out_uptime();
	tags.push_back("<uptime>" + uptime + "</uptime>");

	/* generate random number */
	srand(time(0));
	tags.push_back("<random>" + int2str((rand()%100000)+100) + "</random>");
	
	/* set timestamp */
	tags.push_back("<timeStamp>" + int2str(time(0)) + "</timeStamp>");

	/* set my ip */
	tags.push_back("<ip>" + my_ip + "</ip>");
	// tags.push_back("<ip>127.0.0.1</ip>"); // for debug reasons
	
	if (!shortStatus) {
		tags.push_back("<shortMessage>false</shortMessage>");
		/* append mysql results */
		tags.push_back(tagsForMysqlQueries());
		
		/* append tags for custom commands */
		tags.push_back(tagsForCustomCommands());
		
		/* append tags for sensors */
		tags.push_back(tagsForSensors());
	} else {
		tags.push_back("<shortMessage>true</shortMessage>");
	}
	
	/* mysql status */
	tags.push_back("<mysqlStatus>"+ mysqlStatus +"</mysqlStatus>"); // gets refreshed during tagsForMysqlQueries

	/* set id and pw tags if not authed */
	if (authed == 0) {
		tags.push_back("<id>" + config.smartBeatId() + "</id>");

		tags.push_back("<password>" + config.smartBeatPassword() + "</password>");
	}
	
	std::string message = mix_message(tags);
	// add smartbeat and object tags
	message = "<smartbeat xmlns:os=\"http://liquid-co.de/smartbeat\" xmlns=\"http://liquid-co.de/smartbeat\"><object>"+ message +"</object></smartbeat>";
	// std::cout << message << std::endl;
	/* when message log is enabled then put message to log */
	if (config.logSmartBeatMessages())
		config.logger()->write_to_log("INFO: " + message + "\n");

	message = gpg.encode_message(message);
	if (message == "") {
		config.logger()->write_to_log("ERROR: while gpg encryption. check gpg path in config file e.g. /usr/bin/gpg or try to configure with --enable-64bitworkaround");
		return -1;
	}
	// std::cout << message << std::endl;
	/* put <msg> before and </msg> after encrypted message */
	message = "<msg>" + message + "</msg>";
	
	/* send the stuff */
	if (socket->send(message) == -1) {
		config.logger()->write_to_log("INFO: Status send failed");
		return -1;
	}
	return 0;
}

void Smartbeatclient::write_logmessage_and_sleep(const std::string &message)
{
	config.logger()->write_to_log(message);
	config.logger()->write_to_log("INFO: Restarting");
	socket->close();
	sleep(config.reconnectTime());
}

void Smartbeatclient::nested_loop()
{
	int error;
	std::string message;
	/* nested loop for endless revieving */
	while (1) {
		error = socket->recieve(message);
		if (error == -1) {
			write_logmessage_and_sleep(socket->read_error());
			break;
		}
		
		if (message == "1") {
			error = send_status();
			if ( error == -1) {
				write_logmessage_and_sleep("INFO: could not send status");
				break;
			}
		} else if (message == "0") {
			socket->send("<msg>" + config.clientPinganswer() + "</msg>");
		} else if (message == "2") {
			write_logmessage_and_sleep("INFO: request for short status; check server settings");
			error = send_status(true);
			if ( error == -1) {
				write_logmessage_and_sleep("INFO: could not send status");
				break;
			}
		} else {
			// std::cout << message << std::endl;
			write_logmessage_and_sleep("ERROR: unknown protocol");
			break;
		}
	}
}

void Smartbeatclient::forever()
{
	// int error;
	std::string message;

	/* main loop */
	while (1) {
		socket->close();
		authed = 0; // forget already done authetification
		message = "";
		
		/* error occoured in database */
		if (mysqlReconnect == true && mysqlStatus != "UNUSED") {
			config.logger()->write_to_log("INFO: Reconnecting to Database cause of failures");
			clearDatabases();
			connectToDatabases();
			mysqlReconnect = false;
		}
		/* update my ip */
		if (get_my_ip(my_ip) == -1) {
			write_logmessage_and_sleep("ERROR: while updating ip");
			continue;
		}
		if (socket->connect() == -1) {
			write_logmessage_and_sleep(socket->read_error());
			continue; // begin while loop again
		}
		config.logger()->write_to_log("INFO: Connected successfully to server");
		nested_loop();
		
	}
}

Smartbeatclient::~Smartbeatclient()
{
	clearDatabases();
	socket->close();
	delete socket;
}

std::string Smartbeatclient::tagsForMonitoredDisks() const
{
	const map< string, string > *monitoredDisks = config.monitoredDisks();
	if (monitoredDisks->size() < 1) // got no disks
		return "";
	
	std::string tag = "<monitoredDisks>";
	map<string, string>::const_iterator diskIterator;
	for (diskIterator = monitoredDisks->begin(); diskIterator != monitoredDisks->end(); diskIterator++) {
		tag += "<disk>";
		tag += "<diskName>"+ diskIterator->first +"</diskName>";
		tag += "<diskUsage>"+ find_out_diskspace(diskIterator->second) +"</diskUsage>";
		tag += "</disk>";
		
	}
	tag += "</monitoredDisks>";
	return tag;
}

std::string Smartbeatclient::tagsForMysqlQueries()
{
	const map< string, DatabaseSettings > * databaseSettings = config.databaseSettings();
	if (databases.empty()) { // no active db connections => no queries
		mysqlStatus = "UNUSED";
		return "";
	}
	mysqlStatus = "OK";
	
	bool statementFailed = false;
	std::string queryResult;
	std::string tag = "<mysqlStatements>";
	map< string, DatabaseSettings >::const_iterator dbSettingsIter; // iterator for config
	map<string, DBAccess *>::iterator dbIter; // iterator for databases
	map <string, string>::const_iterator queryIter;
	
	for (dbSettingsIter = databaseSettings->begin(); dbSettingsIter != databaseSettings->end(); dbSettingsIter++) {
		for (queryIter = dbSettingsIter->second.dbQueries.begin();
			queryIter != dbSettingsIter->second.dbQueries.end(); queryIter++) {
				// find database
				dbIter = databases.find(dbSettingsIter->first);
				if (dbIter != databases.end()) {
					if (dbIter->second->query(queryIter->second, queryResult, config.coloumnSeparator(), config.rowSeparator()) == 0) {
						tag += "<mysqlStatement>";
						tag += "<statementIdentifier>"+ queryIter->first +"</statementIdentifier>";
						tag += "<statementResult>"+ replaceSpecialCharacters(queryResult) +"</statementResult>";
						tag += "</mysqlStatement>";
					} else {
						statementFailed = true;
						config.logger()->write_to_log("ERROR: Query \""+ queryIter->second +"\" failed on database "+ dbIter->first);
						config.logger()->write_to_log(dbIter->second->read_error_message());
					}
				}
			}
	}
	tag += "</mysqlStatements>";
	
	if (statementFailed) {
		mysqlStatus = "FAILED";
		mysqlReconnect = true; // trigger reconnect
	}
	
	return tag;
}

std::string Smartbeatclient::tagsForCustomCommands() const
{
	const map< string, string > *customCommands = config.customCommands();
	if (customCommands->size() < 1) // got no disks
		return "";
	
	std::string tag = "<customCommands>";
	map<string, string>::const_iterator commandIterator;
	for (commandIterator = customCommands->begin(); commandIterator != customCommands->end(); commandIterator++) {
		tag += "<command>";
		tag += "<commandIdentifier>"+ commandIterator->first +"</commandIdentifier>";
		std::string cmdResult;
		execute_command(commandIterator->second, cmdResult, MaxCustomCommandResult); // run command
		tag += "<commandResult>"+ replaceSpecialCharacters(cmdResult) +"</commandResult>";
		tag += "</command>";
		
	}
	tag += "</customCommands>";
	return tag;
}

std::string Smartbeatclient::tagsForSensors() const
{
	const map<string, string> *sensors = config.sensorNames();
	if (sensors->size() < 1) // got no sensors
		return "";
	
	boost::regex sensorRegexp("\\W*(.*): *([+-\\w]* ?\\w* ?).*");
	boost::regex sensorRegexpNextLine("\\s*([+-\\w]* ? \\w*).*");
	boost::smatch what;
	boost::smatch whatSensorNextLine;
	
	bool sensorFound = false;
	// fetch the sensor values
	std::string sensorResult;
	execute_command(config.sensorCommand(), sensorResult, MaxCustomCommandResult); // run command
	
	// split sensor results in lines
	vector<string> lines;
	boost::split(lines, sensorResult, boost::is_any_of("\n"));
	
	std::string tag = "<sensors>";
	for (unsigned int i = 0; i < lines.size(); ++i) {
		// std::cout << "line " << lines[i] << std::endl;
		if (boost::regex_match(lines[i], what, sensorRegexp, boost::match_extra)) {
			if (what.size() > 1) {
				for (map<string, string>::const_iterator sensorIter = sensors->begin(); sensorIter != sensors->end(); sensorIter++) {
					if (sensorIter->second == what[1]) {
						sensorFound = true;
						if (what[2] == "" && i < lines.size() - 1) { // value is not in the same line
							if (boost::regex_match(lines[i + 1], whatSensorNextLine, sensorRegexpNextLine, boost::match_extra)) {
								if (whatSensorNextLine.size() > 0) {
									tag += "<sensor>";
									tag += "<sensorName>"+ sensorIter->first +"</sensorName>";
									// just to be sure replace special characters
									tag += "<sensorValue>"+ replaceSpecialCharacters(whatSensorNextLine[1]) +"</sensorValue>"; // add value and unit
									tag += "</sensor>";
								}
							}
						} else {
							tag += "<sensor>";
							tag += "<sensorName>"+ sensorIter->first +"</sensorName>";
							// just to be sure replace special characters
							tag += "<sensorValue>"+ replaceSpecialCharacters(what[2]) +"</sensorValue>"; // add value and unit
							tag += "</sensor>";
						}
					}
				}
			}
		}
	}
	tag += "</sensors>";
	
	if (!sensorFound) // empty tags if no sensor found
		return "";
	
	return tag;
}

std::string Smartbeatclient::replaceSpecialCharacters(const std::string & origXmlContent) const
{
	std::string replacedMessage = origXmlContent;
	
	// replace all special characters
	// but replace only those which could destroy xml
	// all other special characters works as long as
	// gpg can handle them (gpg utf-8 support)
	boost::replace_all(replacedMessage, "<", "&lt;");
	boost::replace_all(replacedMessage, ">", "&gt;");
	boost::replace_all(replacedMessage, "\"", "&quot;");
	boost::replace_all(replacedMessage, "'", "&apos;");
	
	return replacedMessage;
}

